"""
This module contains functions to execute certain Radiance programs with
Python. The Pyrano project is for simulating solar irradiance with Radiance,
therefore the functions here are pre-set with the Radiance parameters to
simulate the full solar spectrum and generate irradiance results (instead of
the visible spectrum with illuminance results).
"""
import subprocess

def _run_command(rad_cmd, args, stdin=None):
    '''rad_cmd=Radiance commande name string, args=list of parameters
    stdin=standard input. Returns a python CompletedProcess object'''
    out = subprocess.run([rad_cmd] + args, stdin=stdin, stdout=subprocess.PIPE)
    return out


def _stdout_to_file(stdout, file_name):
    '''Writes stdout of a subprocess to a file.'''
    # note: newline='\n' so the writen file is unix-readable LF
    with open(file_name , 'w', newline='\n', encoding='UTF-8') as f:
        f.write(stdout.decode(encoding='UTF-8'))


def obj2rad(obj_file, output_rad_file, custom_args=None):
    '''Converts an obj file to a rad file.'''
    rad_cmd = 'obj2rad'
    if custom_args is not None:
        args = custom_args
    else:
        args = ['-f', obj_file]
    output = _run_command(rad_cmd, args, stdin=None)
    if output_rad_file is not None:
        _stdout_to_file(stdout=output.stdout, file_name=output_rad_file)
    else:
        return output.stdout


def oconv(mat_rad_file, geo_rad_file, output_file=None, custom_args=None,
          stdin=None):
    '''Returns the stdout of oconv. Optionally custom arguments can be used
    instead of the predefined material and geometry rad files. If output_file
    is specified, instead of returning it as a python object, the standard
    output will be written to a file.'''
    rad_cmd = 'oconv'
    if custom_args is not None:
        args = custom_args
    else:
        args = [mat_rad_file, geo_rad_file]
    output = _run_command(rad_cmd, args, stdin)
    if output_file is not None:
        with open(output_file , 'wb') as f:
            f.write(output.stdout)
    else:
        return output.stdout


def write_skyglow(file_name, m=1):
    '''Writes a receiver surface rad file as described in:
    "Subramaniam, S., 2017. Daylighting Simulations with Radiance using
    Matrix-based Methods". file_name=name of .rad file to write, m=controls the
    further division of the sky. The basis is the Reinhart
    sky division. It works the same way as MF in radiance: m=1 gives
    the Reinhart sky with 145 sky-patches (the middle of sky segment 1 points
    to 360 degs (North), sky segment 2 is the one next to it in the
    counter-clockwise direction); m=2 divides each
    skypatch to 4 smaller skypatches except for the top one and results in 577
    sky-patches. m=4 results in 2305 sky-patches etc.'''
    with open(file_name , 'w', newline='\n') as f:
        f.write('#@rfluxmtx u=+Y h=u\n')
        f.write('void glow groundglow\n')
        for i in range(2): f.write('0\n')
        f.write('4 1 1 1 0\n')
        f.write('\n')
        f.write('groundglow source ground\n')
        for i in range(2): f.write('0\n')
        f.write('4 0 0 -1 180\n')
        f.write('\n')
        f.write('#@rfluxmtx u=+Y h=r{}\n'.format(m))
        f.write('void glow skyglow\n')
        for i in range(2): f.write('0\n')
        f.write('4 1 1 1 0\n')
        f.write('\n')
        f.write('skyglow source skydome\n')
        for i in range(2): f.write('0\n')
        f.write('4 0 0 1 180\n')


def rfluxmtx(skyglow_file, sp_file, oct_file, lw=0.00001, ab=2, ad=10000,
             n=4, output_file=None, custom_args=None):
    '''Returns the stdout of rfluxmtx. skyglow_file = skyglow.rad file e.g.
    made with the write_skyglow function, sp_file = sensor point file
    containing the  x, y, z, xi, yi, zi, coordinates of the sensorpoints,
    oct_file = octree file made e.g. with the oconv function,
    lw = radiance lw parameter (it is recommended to use lw = (1)/(ad*10),
    ab = radiance ab parameter, ad = radiance ad parameter, n = number of cpu
    cores to use. <-- When these parameters are used, the rfluxmtx is set up to
    provide daylight coefficients as described in "Subramaniam, S., 2017.
    Daylighting Simulations with Radiance using Matrix-based Methods" in
    section 6.1.1.1. This way DC-s are calculated for sensor points as
    "senders". Optionally, for a more general use of rfluxmtx, custom_args can
    be used instead of the predefined ones. In that case all arguments should
    be provided as a list of strings. If output_file is specified, instead of
    returning it as a python object, the standard output will be written to a
    file.'''
    rad_cmd = 'rfluxmtx'
    if custom_args is not None:
        stdin = open(sp_file, 'r')
        args = custom_args
        output = _run_command(rad_cmd, args, stdin)
        stdin.close()
    else:
        with open(sp_file, 'r') as f:
            y = len(f.readlines())
        stdin = open(sp_file, 'r')
        args = ['-I+', '-y', str(y), '-lw', str(lw), '-ab', str(ab), '-ad',
                str(ad), '-n', str(n), '-', skyglow_file, '-i', oct_file]
        output = _run_command(rad_cmd, args, stdin)
        stdin.close()
    if output_file is not None:
        _stdout_to_file(stdout=output.stdout, file_name=output_file)
    else:
        return output.stdout


def epw2wea(epw_file, output_wea_file):
    '''Converts an epw weather file to a wea file.'''
    rad_cmd = 'epw2wea'
    args = [epw_file, output_wea_file]
    _run_command(rad_cmd, args, stdin=None)


def gendaymtx(wea_file, m, g_albedo=0.2, output_file=None, custom_args=None):
    '''Generates sky matrix from a wea file. wea_file = .wea file input,
    m = controls the further division of the sky. The basis is the Reinhart
    sky division. It works the same way as MF in radiance: m=1 gives
    the Reinhart sky with 145 sky-patches; m=2 divides each skypatch to 4 smaller skypatches
    except for the top one and results in 577 sky-patches. m=4 results in 2305
    sky-patches etc. g_albedo=ground albedo, 0 is black, 1 is white.
    output_file = output .smx file name. Optionally, for a more general use of
    rfluxmtx, custom_args can be used instead of the predefined ones. In that
    case all arguments should be provided as a list of strings. If output_file
    is specified, instead of returning it as a python object, the standard
    output will be written to a file.'''
    rad_cmd = 'gendaymtx'
    if custom_args is not None:
        args = custom_args
    else:
        args = ['-m', str(m), '-g', str(g_albedo), str(g_albedo),
                str(g_albedo), '-O1', wea_file]
    output = _run_command(rad_cmd, args, stdin=None)
    if output_file is not None:
        _stdout_to_file(stdout=output.stdout, file_name=output_file)
    else:
        return output.stdout


def calc_irradiance(dc_file, sky_mtx_file, output_file, cr=0.265, cg=0.670,
                    cb=0.065):
    '''This function pipes together dctimestep and rmtxop Radiance programs to
    calculate irradiance output from a sky matrix and daylight coefficients.
    dc_file = daylight coefficient file made with e.g. rfluxmtx function,
    sky_mtx_file = sky matrix made with e.g. gendaymtx function, output_file =
    file to write the irradiance results. cr=0.265, cg=0.670 and cb=0.065 are
    coefficients to convert the 3 RGB channels to solar irradiance.'''
    rad_cmd_dctimestep = 'dctimestep'
    args_dctimestep = [dc_file, sky_mtx_file]
    output_dctimestep = subprocess.Popen([rad_cmd_dctimestep] + args_dctimestep,
                                         stdout=subprocess.PIPE)
    rad_cmd_rmtxop = 'rmtxop'
    args_rmtxop = ['-fa', '-t', '-c', str(cr), str(cg), str(cb), '-']
    output_rmtxop = subprocess.Popen([rad_cmd_rmtxop] + args_rmtxop,
                                     stdin=output_dctimestep.stdout,
                                     stdout=subprocess.PIPE)
    output_dctimestep.stdout.close()
    output = output_rmtxop.communicate()[0]
    output_rmtxop.stdout.close()
    if output_file is not None:
        _stdout_to_file(stdout=output, file_name=output_file)
    else:
        return output



"""
This module contains functions related to reading/writing Radiance-related files.
"""
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
import matplotlib.colors
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
# Pyrano imports
from pyrano.pointcloud import sph2cart
# Geomeppy imports
from geomeppy.view_geometry import _get_collection, _get_collections, _get_limits
from io import StringIO
from eppy.iddcurrent import iddcurrent
from geomeppy import IDF
from geomeppy.geom.polygons import Polygon3D


def plot_sky_segments(sky_segments, dir_arrows=True, return_ax=False,
                      c_values=None, c_v_lims=None, cmap=None, r=1, alpha=0.2):
    '''Plots the sky-segments in 3D from a Pandas DataFrame made with
    divide_sky function. If return_ax is true then instead of plotting the
    function returns a matplotlib.axes object so further plots can be added to
    it. If dir_arrows is True, arrows for the 4 cardinal directions are added
    to the plot. North arrow is red, east arrow is yellow. c_values = iterable
    to color the sky segments (e.g. with DC values) if = None, the plot uses a
    default color; c_v_lims = tuple or list of 2 values for min and max range
    of coloring with c_values; cmap = matplotlib.cm colormapobject to use for
    coloring the sky segments; r = radius of the 3D sky segment plot; alpha =
    transparency of the colors in the plot.'''
    fig = plt.figure(figsize=(6,6))
    ax = Axes3D(fig, proj_type='persp')
    axlim = r/2 + 0.1*r
    ax.set_xlim(-axlim, axlim)
    ax.set_ylim(-axlim, axlim)
    ax.set_zlim(-axlim, axlim)
    # adding polygons of each sky segment to the plot: the vertices
    # (looking from the inside) start from bottom left and go counter clockwise
    if c_values is not None:
        norm = matplotlib.colors.Normalize(c_v_lims[0], c_v_lims[1])
        colors = [cmap(norm(c_value)) for c_value in c_values]
    else:
        colors = ['orange'] + ['deepskyblue'] * (len(sky_segments) - 1)
    sky_segments['colors'] = colors
    for ss in sky_segments.index[1:-1]:
        elevations = [sky_segments.loc[ss, 'elev_low_lims']]*2
        elevations += [sky_segments.loc[ss, 'elev_high_lims']]*2
        azimuths = [sky_segments.loc[ss, 'azim_left_lims']]
        azimuths += [sky_segments.loc[ss, 'azim_right_lims']]*2
        azimuths += [sky_segments.loc[ss, 'azim_left_lims']]
        ss_coords = sph2cart(az_deg=azimuths, el_deg=elevations, r=r)
        verts = [list(zip(ss_coords['x'], ss_coords['y'], ss_coords['z']))]
        coll3d = Poly3DCollection(verts, alpha=alpha,
                         facecolor=sky_segments.loc[ss, 'colors'],
                         edgecolor='k')
        ax.add_collection3d(coll3d)
        # TODO: implement plotting top sky segment later
    # top sky segment:
    top_segments = sky_segments[sky_segments['elev_high_lims'] ==
                   sky_segments.loc[sky_segments.index[-1], 'elev_low_lims']]
    elevations = top_segments['elev_high_lims']
    azimuths = sky_segments['azim_left_lims']
    ss_coords = sph2cart(az_deg=azimuths, el_deg=elevations, r=r)
    verts = [list(zip(ss_coords['x'], ss_coords['y'], ss_coords['z']))]
    coll3d = Poly3DCollection(verts, alpha=alpha,
                facecolor=sky_segments.loc[sky_segments.index[-1], 'colors'],
                edgecolor='k')
    ax.add_collection3d(coll3d)
    if dir_arrows:
        # add North-South arrows
        arrow_azverts = {'n':[345, 15, 0], 's':[165, 195, 180],
                         'w':[255, 285, 270], 'e':[75, 105, 90]}
        arrow_colors = {'n':'r', 's':'b', 'w':'g', 'e':'yellow'}
        for cd in arrow_azverts.keys():
            arrow = sph2cart(az_deg=arrow_azverts[cd], el_deg=[0, 0, 0],
                             r=[1.125*r, 1.125*r, 1.25*r])
            verts = [list(zip(arrow['x'], arrow['y'], arrow['z']))]
            ax.add_collection3d(Poly3DCollection(verts, alpha=0.5,
                                facecolor=arrow_colors[cd], edgecolor='k'))
    if return_ax:
        return ax
    else:
        plt.show(block=False)


def read_rad_dc_file(rad_dc_file):
    '''Returns a Pandas DataFrame from the Radiance-made mtx file'''
    # TODO: Make it more robust. Shouldn't rely on line count [11]
    f = open(rad_dc_file, 'r')
    dclines = [dcline.strip().split('\t') for dcline in f.readlines()[11:]]
    f.close()
    rad_dc_rgb = pd.DataFrame(data=dclines).transpose().astype(float)
    rad_dc = pd.DataFrame()
    for c in rad_dc_rgb.columns:
        rad_dc[c] = [np.mean((rad_dc_rgb.loc[i, c], rad_dc_rgb.loc[i+1, c],
        rad_dc_rgb.loc[i+2, c])) for i in range(0, len(rad_dc_rgb), 3)]
    return rad_dc


def read_smx_file(smx_file, index=None):
    '''Returns a Pandas DataFrame from the .smx file'''
    # this is very slow for now
    f = open(smx_file, 'r')
    flines = f.readlines()
    for l in flines:
        if l.startswith('NROWS'):
            nskypatch = int(l.split('=')[1])
        if l.startswith('NCOLS'):
            ntimestep = int(l.split('=')[1])
            break
    f.close()
    flines = flines[8:] # TODO: make this more robust later
    smx_df = pd.DataFrame()
    ss = 0
    ill_avgs = []
    for l in flines:
        if l == '\n':
            smx_df[ss] = ill_avgs
            ill_avgs = []
            ss += 1
        else:
            lums = l.strip().split()
            ill_avgs.append(np.mean(np.array(lums, dtype=float)))
        if ss > nskypatch:
            break
    if index is not None:
        smx_df.index = index
    return smx_df


def read_ill_file(ill_file, index=None, colnames=None, skiprows=15):
    '''Returns a Pandas DataFrame from the .ill file (the output file of a
    Radiance simulation). index: to replace the index column (e.g. with the
    timestamps of teh weather file used for the simulation). colnames: to
    replace the column names in the returned DataFrame. skiprows: number of
    rows to skip at the beginning of the ill file. The default 15 should work
    as of the current version of Radiance'''
    f = open(ill_file, 'r')
    nrows, ncols = None, None
    rl = 0
    while nrows == None or ncols == None:
        l = f.readline()
        rl += 1
        if l.startswith('NROWS'):
            nrows = int(l.split('=')[1])
        if l.startswith('NCOLS'):
            ncols = int(l.split('=')[1])
    ills_df = pd.read_csv(f, sep='\t', skiprows=skiprows-rl, header=None)
    f.close()
    del ills_df[ills_df.columns[-1]]
    # doublecheck if all rows and cols are read
    if (nrows, ncols) != ills_df.shape:
        raise ValueError('Mismatch between nrows or ncols in the ill file and '
                         'the returned DataFrame: '
                         'Check skiprows parameter and change if necessary.')
    # replacing index and column names if they are provided
    if index is not None:
        ills_df.index = index
    if colnames is not None:
        ills_df.columns = colnames
    return ills_df


def view_idf_to_ax(fname=None):
    """This is originally from https://github.com/jamiebull1/geomeppy/blob/master/geomeppy/view_geometry.py
    This just returns an ax instead of viewing it on order to  plot it together with the sensorpoints"""
    # set the IDD for the version of EnergyPlus
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() is None:
        IDF.setiddname(iddfhandle)
    # import the IDF
    idf = IDF(fname)
    # create the figure and add the surfaces
    ax = plt.axes(projection="3d")
    collections = _get_collections(idf, opacity=0.5)
    for c in collections:
        ax.add_collection3d(c)
    # calculate and set the axis limits
    limits = _get_limits(idf=idf)
    ax.set_xlim(limits["x"])
    ax.set_ylim(limits["y"])
    ax.set_zlim(limits["z"])
    return ax


def view_idf_and_ill(idf_name, sps, ill, return_ax=False, vmin=0, vmax=1000):
    """To view the e+ IDF and the sensorpoints together. Ill should be a
    pd.series with time in index"""
    surfcoords = []
    polys = []
    for surf in sps:
        polys.append(Polygon3D(surf['surf_coords']))
        for sp in surf['sensor_points']:
            surfcoords.append((sp[0], sp[1], sp[2]))
    xs = [c[0] for c in surfcoords]
    ys = [c[1] for c in surfcoords]
    zs = [c[2] for c in surfcoords]
    ax = view_idf_to_ax(fname=idf_name)
    ax.scatter(xs, ys, zs, marker='o', s=2, c=ill, cmap='nipy_spectral',
               vmin=vmin, vmax=vmax)
    if return_ax:
        return ax
    else:
        plt.show(block=False)



'''This module contains functions to generate sensor points for Radiance
simulations and to work with EnergyPlus IDF geometry.
Some (most) of the functions here are from the Daypym project:
https://gitlab.tue.nl/bp-tue/daypym'''

import numpy as np
import pandas as pd
import json
from matplotlib import pyplot as plt
from geomeppy.geom.transformations import align_face, invert_align_face
from geomeppy import IDF, view_geometry
from geomeppy.geom.polygons import (break_polygons, Polygon2D, Polygon3D,
                                    Vector2D, Vector3D)
from geomeppy.view_geometry import (_get_collection, _get_collections,
                                    _get_surfaces, _get_limits)
from io import StringIO
from eppy.iddcurrent import iddcurrent

def IDFsurf_to_poly(surface):
    "Makes a poly from IDF surface"
    # get its vertices
    corstrings = ['_Xcoordinate', '_Ycoordinate', '_Zcoordinate']
    vertnums = list(range(1, int(surface['Number_of_Vertices'])+1, 1))
    surf_verts = []
    for vn in vertnums:
        vert_cords = tuple([surface['Vertex_'+ str(vn) + cs] for cs in corstrings])
        surf_verts.append(vert_cords)
    # make a 3d polygon from vertices
    poly = Polygon3D(surf_verts)
    return poly


def grid_2d(bbox_2d, d, edges):
    """Creates a grid over the bbox with siden divided to n. TODO: n-ratio"""
    a = abs((bbox_2d[0] - bbox_2d[2])[0])
    b = abs((bbox_2d[0] - bbox_2d[2])[1])
    n = d * a * b
    y1 = (-(a/b + 1) + np.sqrt((a/b + 1)**2 - 4*a/b*(1 - a*b*d))) / (2 * a/b)
    y2 = (-(a/b + 1) - np.sqrt((a/b + 1)**2 - 4*a/b*(1 - a*b*d))) / (2 * a/b)
    x1 = y1* a/b
    x2 = y2* a/b
    y = max(y1, y2)
    x = y * a/b
    #print('a={}; b={}; x={}; y={}; n={}'.format(a,b,x,y,n))
    xs = (bbox_2d.vertices_list[0][0], bbox_2d.vertices_list[2][0])
    ys = (bbox_2d.vertices_list[0][1], bbox_2d.vertices_list[2][1])
    xv = np.linspace(xs[0], xs[1], round(x+1, 0))
    yv = np.linspace(ys[0], ys[1], round(y+1, 0)) # this should be a func of the reatio of width height of the bbox
    grid = np.meshgrid(xv, yv)
    if edges == False:
        lr = max(grid[0].max() - grid[0].min(), grid[1].max() - grid[1].min()) # calc the length of the module in the row-direction (longer direction)
        lc = min(grid[0].max() - grid[0].min(), grid[1].max() - grid[1].min()) # calc the length of the module in the column-direction (shorter direction)
        n_row = grid[0].shape[0]
        n_col = grid[0].shape[1]
        dr = (lr/n_row) / 2
        dc = (lc/n_col) / 2
        grid_transp = [grid[0] + dr, grid[1] + dc] # Pushing the grid to the center of the cells. Now we have a row and column of points off the surface of the pv. These are needed to be removed with gridpoints_in_poly_2d func!
        return grid_transp
    elif edges == True:
        return grid
    else:
        raise ValueError('"edges" should be = "True" or "False"')


def grid_2d_pvmodule(n_row, n_col, bbox_2d):
    """This makes a rectangular grid over the pv cells. Use the
    gridpoints_in_poly_2d on the list of arrays it returns, because this
    produces some extra points as well, that are needed to be clipped off.
    Define rows and cols as n_row > n_col!"""
    #TODO: Make it unambiguous what happens when n_row = n_col
    a = abs((bbox_2d[0] - bbox_2d[2])[0])
    b = abs((bbox_2d[0] - bbox_2d[2])[1])
    area = a * b
    d = ((n_col*n_row) + (n_row + 1) + (n_col + 1) - 1) / area # calc the point density to have 1 sp over each cell
    test_grid = grid_2d(bbox_2d=bbox_2d, d=d, edges=True) # make a 2D test grid on the x-y plane (daypym) This grid contains the edges as well. Not good yet. We will push this with half cell-size in eac direction.
    lr = max(test_grid[0].max() - test_grid[0].min(), test_grid[1].max() -
             test_grid[1].min()) # calc the length of the module in the row-direction (longer direction)
    lc = min(test_grid[0].max() - test_grid[0].min(), test_grid[1].max() -
             test_grid[1].min()) # calc the length of the module in the column-direction (shorter direction)
    dr = (lr/n_row) / 2
    dc = (lc/n_col) / 2
    test_grid_transp = [test_grid[0] + dr, test_grid[1] + dc] # Pushing the grid to the center of the cells. Now we have a row and column of points off the surface of the pv. These are needed to be removed with gridpoints_in_poly_2d func!
    return test_grid_transp


def point_in_poly(point, poly):
    "True if a 2D point is in or on the edge of a 2D poly."
    # TODO: add dynamic quasi infinite point distance with bounding box clue
    qip1_r = Vector2D(10000, 0) # quasi-infinite point 1 to the right
    qip2_r = Vector2D(10000, 1) # quasi-infinite point 2 to the right
    qip1_l = Vector2D(-10000, 0) # quasi-infinite point 1 to the left
    qip2_l = Vector2D(-10000, 1) # quasi-infinite point 2 to the left
    qip1_u = Vector2D(0, 10000) # quasi-infinite point 1 upwards
    qip2_u = Vector2D(1, 10000) # quasi-infinite point 2 upwards
    qip1_d = Vector2D(0, -10000) # quasi-infinite point 1 downwards
    qip2_d = Vector2D(1, -10000) # quasi-infinite point 2 downwards
    test_poly_r = Polygon2D([point, qip1_r, qip2_r])#.order_points('upperleftcorner') # maybe we wouldn't need this anyway
    test_poly_l = Polygon2D([point, qip1_l, qip2_l])#.order_points('upperleftcorner') # maybe we wouldn't need this anyway
    test_poly_u = Polygon2D([point, qip1_u, qip2_u])
    test_poly_d = Polygon2D([point, qip1_d, qip2_d])
    #poly = poly.order_points('upperleftcorner') # maybe we wouldn't need this anyway
    inter_r = poly.intersect(test_poly_r)
    inter_l = poly.intersect(test_poly_l)
    inter_u = poly.intersect(test_poly_u)
    inter_d = poly.intersect(test_poly_d)
    if len(inter_r) == 0:
        inter_r.append([0])
    if len(inter_l) == 0:
        inter_l.append([0])
    if len(inter_u) == 0:
            inter_u.append([0])
    if len(inter_d) == 0:
            inter_d.append([0])
    no_inter_r = len(inter_r[0])
    no_inter_l = len(inter_l[0])
    no_inter_u = len(inter_u[0])
    no_inter_d = len(inter_d[0])
    return True if no_inter_r == 3 or no_inter_l == 3 or no_inter_u == 3 or no_inter_d == 3 else False # we will see how robust this is # check this again later


def gridpoints_in_poly_2d(grid_2d, poly_2d):
    """Returns a list of points, that are inside a 2d poly. Takes a meshgrid
    (made with grid_2d) and a 2d poly as input"""
    pip = [] # list of points in poly
    for x, y in list(zip(grid_2d[0], grid_2d[1])):
        for cx, cy in list(zip(x, y)):
            test_point = Vector2D(cx, cy)
            if point_in_poly(point=test_point, poly=poly_2d):
                pip.append(test_point)
    return pip


def pos_in_module(n_row, n_col, test_grid_transp):
    """Returns a tuple of the local column and row and column (ri, ci) indexes
    within the PV module"""
    mod_pos = []
    if test_grid_transp[0].shape != (n_row+1, n_col+1):
        for c in range(n_col):
            mod_pos.append([(r, c) for r in range(n_row)])
    else:
        for r in range(n_row):
            mod_pos.append([(r, c) for c in range(n_col)])
    ri = []
    for row in mod_pos:
        ri = ri + [col[0] for col in row]
    ci = []
    for row in mod_pos:
        ci = ci + [col[1] for col in row]
    return ri, ci


def create_sensor_points(surf_name, points_in_poly_2d, row_index, col_index,
                         original_poly, sp_offset=0.01, sp_pos_round=3):
    """Returns a dict {'surf_name':surface name, 'sensor_points':list of sensor
    points, 'ri':row_index, 'ci':col_index, 'sp_ori':list of normal vectors}.
     sesor points are translated with sp_offset from the original plane. Takes
     points_in_poly_2d (made with gridpoints_in_poly_2d function) and the
     original poly as input. We can round the coords of the sps. Make sure,
     that the rounding has more digits, than the smallest distances and the
     offset. Default is 1 cm offset and 1 mm rounding."""
    # make a 3D polygon from these points and rotate them back
    sensor_points_poly_trans = Polygon3D(points_in_poly_2d)
    sensor_points_poly = invert_align_face(original=original_poly,
                                           poly2=sensor_points_poly_trans)
    normal_vectors = np.array([original_poly.normal_vector[0],
                              original_poly.normal_vector[1],
                              original_poly.normal_vector[2]])
    transl_vect = normal_vectors * sp_offset
    sensor_points = [list(np.add(tuple([round(sp[i], sp_pos_round) for i in range(len(sp))]), transl_vect)) for sp in sensor_points_poly.vertices_list] # sorry
    sensor_points_surf = {'surf_name':surf_name,
                          'surf_coords':original_poly.vertices_list,
                          'sensor_points':sensor_points,
                          'ri':row_index,
                          'ci':col_index,
                          'sp_ori':list(normal_vectors)} # make a dict for the sensor points with name points and vectors, later export this to DS file
    return sensor_points_surf


def xyz_coords_from_sps(sps):
    """Takes sps dict as input. Extracts x, y, z coords from an sps dict made
    with the create_sensor_points() or the pointgrid_over_surface() function"""
    x, y, z = ([] for i in list(range(3)))
    for p in sps:
        x += [c[0] for c in p['sensor_points']]
        y += [c[1] for c in p['sensor_points']]
        z += [c[2] for c in p['sensor_points']]
    return pd.DataFrame(data={'x':x, 'y':y, 'z':z})


def translate_to_ds_pts(surf_sensor_points, p_name, grid_method, fname=None):
    """Translates and saves the sensorpoints as a .pts file.
    Inputs: grid_method: use grid_method = "pv_module" for saving sensor point
    metadata to simulate irradiance for PVMismatch simulations.
    Use grid_method = "point_density" for saving metadata for general
    irradiance simulation purposes.
    Also saves a csv or json with additional info useful for postprocessing.
    If fname != the sps will be saved to the given location not the default pts
    folder"""
    x, y, z, vx, vy, vz, surf_name, ri, ci = ([] for i in list(range(9)))
    for surf in surf_sensor_points:
        for i in list(range(len(surf['sensor_points']))):
            x.append(surf['sensor_points'][i][0])
            y.append(surf['sensor_points'][i][1])
            z.append(surf['sensor_points'][i][2])
            vx.append(surf['sp_ori'][0])
            vy.append(surf['sp_ori'][1])
            vz.append(surf['sp_ori'][2])
            surf_name.append(surf['surf_name'])
            if grid_method == 'pv_module':
                ri.append(surf['ri'][i])
                ci.append(surf['ci'][i])
                spdf = pd.DataFrame(data={'x':x, 'y':y, 'z':z, 'vx':vx,
                                          'vy':vy, 'vz':vz, 'ri':ri, 'ci':ci,
                                          'surf_name':surf_name})
            elif grid_method == 'point_density':
                spdf = pd.DataFrame(data={'x':x, 'y':y, 'z':z, 'vx':vx,
                                          'vy':vy, 'vz':vz,
                                          'surf_name':surf_name})
            else:
                raise ValueError('"grid_method" should be = "pv_module" or "point_density"')
    spdf.index.name = 'sp_index'
    if fname == None:
        savepath = r'pts/'
        save_fname = p_name
    else:
        savepath = r''
        save_fname = fname
    ptsfile = open(r'{}{}.pts'.format(savepath, save_fname), 'w')
    for p in spdf.index:
        ptsfile.write(str(spdf['x'][p]) +' '+ str(spdf['y'][p]) +' '+
                      str(spdf['z'][p]) +' '+ str(spdf['vx'][p]) +' '+
                      str(spdf['vy'][p]) +' '+ str(spdf['vz'][p]) + '\n')
    ptsfile.close()
    # saving metadatata of sensor points in csv or json format
    spdf.to_csv(r'{}{}.csv'.format(savepath, save_fname))
    for s in surf_sensor_points:
        spi = spdf[spdf['surf_name']==s['surf_name']].index.tolist()
        s.update({'sp_index':[int(i) for i in spi]})
    with open(r'{}{}.json'.format(savepath, save_fname), 'w') as fp:
        json.dump(surf_sensor_points, fp, indent=4)
    return


def view_idf_to_ax(fname=None, idf_txt=None, test=False):
    """This is originally from https://github.com/jamiebull1/geomeppy/blob/master/geomeppy/view_geometry.py
    This just returns an ax instead of viewing it on order to  plot it together
    with the sensorpoints"""
    # type: (Optional[str], Optional[str], Optional[bool]) -> None
    if fname and idf_txt:
        raise ValueError("Pass either fname or idf_txt, not both.")
    # set the IDD for the version of EnergyPlus
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() is None:
        IDF.setiddname(iddfhandle)
    if fname:
        # import the IDF
        idf = IDF(fname)
    elif idf_txt:
        idf = IDF()
        idf.initreadtxt(idf_txt)
    # create the figure and add the surfaces
    ax = plt.axes(projection="3d")
    collections = _get_collections(idf, opacity=0.5)
    for c in collections:
        ax.add_collection3d(c)
    # calculate and set the axis limits
    limits = _get_limits(idf=idf)
    ax.set_xlim(limits["x"])
    ax.set_ylim(limits["y"])
    ax.set_zlim(limits["z"])
    return ax


def view_idf_and_sps(p_name, idf_name, sps, savefig=False, return_ax=False):
    """To view the e+ IDF and the DS sensorpoints together. Utilizing modified
    version of Geomeppy view_idf function:
    https://github.com/jamiebull1/geomeppy/"""
    # TODO implement auto save fig: Save_fig False: no saving, True: save it to /geo
    surfcoords = []
    polys = []
    for surf in sps:
        polys.append(Polygon3D(surf['surf_coords']))
        for sp in surf['sensor_points']:
            surfcoords.append((sp[0], sp[1], sp[2]))
    xs = [c[0] for c in surfcoords]
    ys = [c[1] for c in surfcoords]
    zs = [c[2] for c in surfcoords]
    ax2 = view_idf_to_ax(fname=idf_name, idf_txt=None, test=False)
    ax2.scatter(xs, ys, zs, marker='o', s=2, c='k')
    if savefig:
        try:
            plt.savefig('geo/{}.png'.format(p_name))
        except:
            plt.savefig('{}.png'.format(p_name))
    if return_ax:
        return ax2
    else:
        plt.show(block=False)


def write_manual_sp(pts_file, x, y, z, vx, vy, vz, append=False):
    '''Writes a pts sensor point file. x, y, z are the coordinates of the point
    and vx, vy, vz are the coordinates of the direction vector of the point.
    If append = True, a point will be added to teh pts file in a new line, else
    the pts file is overriden.'''
    if append:
        with open(pts_file, 'a' , newline='\n') as f:
            f.write('{} {} {} {} {} {}\n'.format(x, y, z, vx, vy, vz))
    else:
        with open(pts_file, 'w' , newline='\n') as f:
            f.write('{} {} {} {} {} {}\n'.format(x, y, z, vx, vy, vz))


def pointgrid_over_surface(surface, d=5, edges=True, sp_offset=0.01):
    """Wrapper function for create_sensor_points(). Inputs:
    surface: IDF urface (via geomeppy), d: desired point density sps/m2
    edges: bool to decide if we want poits on the edge of the surface,
    sp_offset: offset of sensorpoints from the surface in m. Returns a dict:
    {'surf_name':surface name, 'sensor_points':list of sensor points,
    'ri':row_index, 'ci':col_index, 'sp_ori':list of normal vectors}."""
    # make poly from idf (daypym)
    poly = IDFsurf_to_poly(surface=surface)
    # translate poly to the x-y plane (geomeppy)
    poly_trans = align_face(poly).order_points('upperleftcorner')
    # make the poly 2D (geomeppy)
    poly_trans_2d = poly_trans.project_to_2D()
    # make a 2d bbox of the translated surface on the xy pane (geomeppy)
    poly_trans_bbox_2d = poly_trans.bounding_box.project_to_2D()
    # make a 2D test grid on the x-y plane (daypym)
    test_grid = grid_2d(bbox_2d=poly_trans_bbox_2d, d=d, edges=edges)
    # makes a list of points, that are inside the 2D poly (daypym)
    pip = gridpoints_in_poly_2d(grid_2d=test_grid, poly_2d=poly_trans_2d)
    # sensor point data dict (daypym)
    sp = create_sensor_points(surf_name=surface['Name'], points_in_poly_2d=pip,
    row_index=[0], col_index=[0], original_poly=poly, sp_offset=sp_offset,
    sp_pos_round=3)
    return sp


def pointgrid_over_pvmodule(surface, n_row=10, n_col=6, sp_offset=0.01):
    """Wrapper function to create sensot points over each cell of a pv module.
    Inputs: surface: IDF urface (via geomeppy), n_row: number of cell-rows in
    the PV module n_col: number of cell-columns in the PV module
    sp_offset: offset of sensorpoints from the surface in m. Returns a dict:
    {'surf_name':surface name, 'sensor_points':list of sensor points,
    'ri':row_index, 'ci':col_index, 'sp_ori':list of normal vectors}."""
    # make poly from idf (daypym)
    poly = IDFsurf_to_poly(surface=surface)
    # translate poly to the x-y plane (geomeppy)
    poly_trans = align_face(poly).order_points('upperleftcorner')
    # make the poly 2D (geomeppy)
    poly_trans_2d = poly_trans.project_to_2D()
    # make a 2d bbox of the translated surface on the xy pane (geomeppy)
    poly_trans_bbox_2d = poly_trans.bounding_box.project_to_2D()
    # make a 2D test grid on the x-y plane (daypym)
    test_grid = grid_2d_pvmodule(n_row=n_row, n_col=n_col,
                                 bbox_2d=poly_trans_bbox_2d)
    # makes a list of points, that are inside the 2D poly (daypym)
    pip = gridpoints_in_poly_2d(grid_2d=test_grid, poly_2d=poly_trans_2d)
    # row and column index information
    ri, ci = pos_in_module(n_row=n_row, n_col=n_col, test_grid_transp=test_grid)
    # sensor point data dict (daypym)
    sp = create_sensor_points(surf_name=surface['Name'], points_in_poly_2d=pip,
    row_index=ri, col_index=ci, original_poly=poly, sp_offset=sp_offset,
    sp_pos_round=3)
    return sp



"""
This module contains functions related to manipulating LiDAR point clouds
to use them for calculating shading for hourly annual solar irradiance
simulations.
Some of the functions are from the USM_LiDAR project:
https://gitlab.tue.nl/bp-tue/usm_lidar_dev
and the Daypym project:
https://gitlab.tue.nl/bp-tue/daypym
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import time
from mpl_toolkits.axes_grid1 import make_axes_locatable
from PIL import Image
from geomeppy.geom.polygons import (Polygon2D, Vector2D)
from pyrano.geometry import point_in_poly
import rtree


def sph2cart(az_deg, el_deg, r):
    """Function to convert spherical coordinates to cartesian coordinates.
    It expects numpy arrays in degrees as input with the same length or length
    of 1"""
    # convert to radians
    az_rad = np.deg2rad(az_deg)
    el_rad = np.deg2rad(el_deg)
    # do the maths
    rcos_theta = r * np.cos(el_rad)
    x = rcos_theta * np.cos(az_rad)
    y = rcos_theta * np.sin(az_rad)
    z = r * np.sin(el_rad)
    # put the result into a DataFrame
    cart = pd.DataFrame({'x': x, 'y': y, 'z': z})
    return cart


def cart2sph(x, y, z):
    """Function to convert cartesian coordinates to spherical coordinates.
    It expects numpy arrays as input with the same length or length of 1"""
    # do the maths
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    el_rad = np.arccos(z/r)
    az_rad = np.arctan2(y, x)
    # convert to degrees and rotate 90 degrees to have North be 0 degrees
    el_deg = 90 - np.rad2deg(el_rad)  # 90-x to get elevation instead of zenith
    az_deg = -np.rad2deg(az_rad) + 90
    # wrap around the parts above 360 due to rotate
    az_deg[az_deg < 0] = az_deg[az_deg < 0] + 360
    # put the result into a DataFrame
    sph = pd.DataFrame({'azimuth': az_deg, 'elevation': el_deg, 'r': r})
    return sph


def supress_points(xyz, vertices, zmax, apothem):
    """Limits an xyz point cloud's z coordinates to zmax within the boundaries
    of a polygon. xyz: a DataFrame with x-y-z columns made by teh tiff2xyz()
    function; vertices: vertices of the polygon; zmax: max height of points
    within the polygon; aphotem: aphotem used in the tiff2xyz function()."""
    vert_vects = [Vector2D(v[0]-apothem, v[1]-apothem) for v in vertices]
    poly = Polygon2D(vert_vects)
    x_min = min(poly.xs)
    x_max = max(poly.xs)
    y_min = min(poly.ys)
    y_max = max(poly.ys)
    # sorting from high z to low to speed things up
    for i in xyz.sort_values(by='z', ascending=False).index:
        if ((xyz.loc[i, 'x'] > x_min and xyz.loc[i, 'x'] < x_max) and
           (xyz.loc[i, 'y'] > y_min and xyz.loc[i, 'y'] < y_max)):
           point = Vector2D(xyz.loc[i, 'x'], xyz.loc[i, 'y'])
           if point_in_poly(point, poly):
               xyz.loc[i, 'z'] = zmax
        if xyz.loc[i, 'z'] < zmax:
            break
    return xyz


def shift_cart_coords(xyz, p_xyz_ref, p_xyz_to_shift):
    """Shifts xyz coordinates to a new reference. Inputs: xyz = the input xyz
    DataFrame to be shifted, p_xyz_ref = xyz coordinates dict of a reference
    point in the ref_xyz e.g. {'x':0, 'y':0, 'z':0}, p_xyz_shfted = xyz coords
    of the same reference point in the to be shifted xyz. It returns a new xyz
    that is shifted to the reference system."""
    p_delta = {'x':p_xyz_to_shift['x'] - p_xyz_ref['x'],
               'y':p_xyz_to_shift['y'] - p_xyz_ref['y'],
               'z':p_xyz_to_shift['z'] - p_xyz_ref['z']}
    for c in ['x', 'y', 'z']:
        xyz[c] -= p_delta[c]
    return xyz


def lidar_to_ep_coords(lidar_xyz, ref_point_ep, ref_point_lidar):
    '''Rerurns the xyz DataFrame of the LiDAR point cloud in the coordinate
    system of the energy plus model. lidar_xyz = xyz DataFrame of the LiDAR
    point cloud. ref_point_ep = list of x, y, z coordinates of a reference
    point in the EnergyPlus model, ref_point_lidar = list of x, y, z
    coordinates of the same reference point in the LiDAR point cloud.'''
    apothem = -lidar_xyz['y'].min()
    p_xyz_ref = {'x':ref_point_ep[0], 'y':ref_point_ep[1], 'z':ref_point_ep[2]}
    p_xyz_to_shift = {'x':ref_point_lidar[0], 'y':ref_point_lidar[1],
                      'z':ref_point_lidar[2]}
    xyz_ep = shift_cart_coords(lidar_xyz, p_xyz_ref, p_xyz_to_shift)
    xyz_ep['x'] += apothem
    xyz_ep['y'] += apothem
    return xyz_ep


def tiff2array(tiffloc):
    """function to read tiff file and put it into an array"""
    # set maximum number of pixels higher to prevent DOS errors
    Image.MAX_IMAGE_PIXELS = 125000000
    # open tiff file and convert it to an array
    array = np.array(Image.open(tiffloc))
    # get the maximum value of the whole array
    nanpoint = np.amax(array)
    # set the maximum value to be nan
    array[array == nanpoint] = np.nan
    # flip array around the x-axis
    array = np.flip(array, axis=0)
    return array


def lidarzoom(array, x, y, apothem=200, resolution=0.5):
    """function that 'zooms in' to a test point (x, y) of the LiDAR array
        and removes all unnecessary points outside of the apothem (square
        radius)
    array = the LiDAR array to 'zoom in' to
    x = x coordinate of the test point in the LiDAR array
    y = y coordinate of the test point in the LiDAR array
    apothem = apothem ('square radius') of LiDAR points that need to be kept
    resolution = resolution of the tiff file (0.5 for pdok.nl files)"""
    # multiply x, y and apothem by 1/resolution to get distance in meters
    x = int(x * (1/resolution))
    y = int(y * (1/resolution))
    apothem = int(apothem * (1/resolution))
    xrangeleft = x - apothem
    xrangeright = x + apothem
    yrangebottom = y - apothem
    yrangetop = y + apothem
    height, width = array.shape
    array = np.delete(array, np.r_[0:xrangeleft, xrangeright:width], axis=1)
    array = np.delete(array, np.r_[0:yrangebottom, yrangetop:height], axis=0)
    return array


def array2xyz_depr1(array, savename='pointcloud', save=False, resolution=0.5):
    """function to make the LiDAR array into an XYZ array and save an XYZ
    file"""
    # make array into xyz array
    height, width = array.shape
    ys, xs = np.mgrid[:height/2:resolution, :width/2:resolution]
    xyz = np.column_stack((xs.ravel() - round(width/2/(1/resolution)),
                           ys.ravel() - round(height/2/(1/resolution)),
                           array.ravel()))
    # remove rows with NaNs
    xyz = xyz[~np.isnan(xyz).any(axis=1)]
    # convert array to DataFrame for easier use
    xyz = pd.DataFrame({'x': xyz[:, 0], 'y': xyz[:, 1], 'z': xyz[:, 2]})
    # save the XYZ file
    if save:
        save_df(xyz, filename=savename, filetype='xyz', folder='Pointclouds',
                header=False, sep=' ')
    return xyz


def array2xyz(array, savename='pointcloud', save=False, resolution=0.5):
    """function to make the LiDAR array into an XYZ array and save an XYZ
    file"""
    # make array into xyz array
    height, width = array.shape
    ys, xs = np.mgrid[:height*resolution:resolution, :width*resolution:resolution]
    xyz = np.column_stack((xs.ravel() - round(width/2/(1/resolution)),
                           ys.ravel() - round(height/2/(1/resolution)),
                           array.ravel()))
    # remove rows with NaNs
    xyz = xyz[~np.isnan(xyz).any(axis=1)]
    # convert array to DataFrame for easier use
    xyz = pd.DataFrame({'x': xyz[:, 0], 'y': xyz[:, 1], 'z': xyz[:, 2]})
    # save the XYZ file
    if save:
        save_df(xyz, filename=savename, filetype='xyz', folder='Pointclouds',
                header=False, sep=' ')
    return xyz


def lidarplot(array, figure_name=None, resolution=0.5):
    """function to plot lidar maps based on the LiDAR array
    save = boolean whether to save the plot
    resolution = resolution of the tiff file (0.5 for pdok.nl files)"""
    # initialize figure
    fig, ax = plt.subplots()
    fig.tight_layout()
    fig.set_size_inches(5, 5)
    # set title and axes labels
    ax.set_title('DSM')
    ax.set_xlabel('X-axis [m]')
    ax.set_ylabel('Y-axis [m]')
    ax.minorticks_on()
    # get shape to put into imshow extent
    m, n = array.shape
    # create figure
    # extend is divided by 1/resolution to go to meters on the axes
    im = plt.imshow(array, cmap='plasma', origin='lower', interpolation='none',
                    extent=[0, n/(1/resolution), 0, m/(1/resolution)],
                    aspect='equal')
    # add color bar to figure
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size=0.2, pad=0.1)
    cbar = plt.colorbar(im, cax=cax, extend='both')
    cbar.minorticks_on()
    cbar.set_label(label='Height above Amsterdam Ordnance Datum [m]')
    plt.show(block=False)
    if figure_name is not None:
        plt.save_figure(fig, figure_name, filetype='png')


def tiff2xyz(tiffloc, x=None, y=None, apothem=200, save_xyz_filename=None,
             plot_tiff=False, plot_croppedtiff=False, resolution=0.5):
    """Function to go from the tiff file, with an x and y in that file around
    the testpoint to an xyz-file with the ability to plot it and return the
    x, y, z array
    tiffloc = relative or full file location of the tiff file (e.g. downloaded
        from pdok.nl)
    x = x coordinate of the test point in the full tiff file
    y = y coordinate of the test point in the full tiff file
    apothem = 'square radius' you want to take into account in the cropped
        tiff version
    savexyz = boolean whether to save the .xyz-file
    plot_tiff = boolean whether to plot the full tiff file
    plot_croppedtiff = boolean whether to plot the cropped tiff version
    resolution = resolution of the tiff file (0.5 for pdok.nl files)."""
    # load tiff file in an array
    array = tiff2array(tiffloc)
    # plot full tiff file
    if plot_tiff:
        lidarplot(array, figure_name=None, resolution=resolution)
    # crop array around x, y with a radius
    if x and y and apothem is not None:
        array = lidarzoom(array, x, y, apothem, resolution=resolution)
    # plot cropped array
    if plot_croppedtiff:
        lidarplot(array, figure_name=None, resolution=resolution)
    # convert array to x, y and z DataFrame
    xyz = array2xyz(array, savename=None, save=False, resolution=resolution)
    if save_xyz_filename is not None:
        xyz.to_csv(save_xyz_filename, sep=' ', header=False, index=False)
    return xyz


def dist2D(p, q):
    """Return the Euclidean distance between 2D points p and q."""
    return np.hypot(p[0] - q[0], p[1] - q[1])


def remove_neighbors(points, tol):
    """Function to remove nearest neighbors of 2D points collection DataFrame
        within tolerance and returns 'culled' DataFrame"""
    # create an empty list so we can fill it
    result = []
    # create the R-tree
    index = rtree.index.Index()
    # get the column names of the points DataFrame
    columns = points.columns.values.tolist()
    # put the points in a list
    points = points.values.tolist()
    # iterate over points and put points that comply with conditions in the
    # result list
    for i, p in enumerate(points):
        # get x and y value of the points
        x, y = p
        # get the proximity of the point to the others
        nearby = index.intersection((x - tol, y - tol, x + tol, y + tol))
        # keep the points that are above the tolerance in the proximity set
        if all(dist2D(p, points[j]) >= tol for j in nearby):
            # append the point to the result list
            result.append(p)
            # insert the point in the index
            index.insert(i, (x, y, x, y))
    # create a dataframe of the result list with the original column names
    result = pd.DataFrame(result, columns=columns)
    return result


def plotprojected(projected, save=False, radius=10):
    """Plot DataFrame with x, y and z column in 3D"""
    # set up plot
    fig = plt.figure()
    ax = Axes3D(fig)
    ax.set_xlabel('X axis')
    ax.set_ylabel('Y axis')
    ax.set_zlabel('Z axis')
    ax.set_xlim3d(-radius, radius)
    ax.set_ylim3d(-radius, radius)
    # plot x, y and z of the DataFrame
    ax.scatter3D(projected['x'], projected['y'], projected['z'])
    # save figure
    if save:
        save_figure(fig, figurename='projected')


def cyl2cart(az_deg, z, rho):
    """function to convert cylindrical coordinates to cartesional coordinates.
    It expects numpy arrays as input with the same length or length of 1"""
    # convert to radians
    az_rad = np.deg2rad(az_deg)
    # do the maths
    x = rho * np.cos(az_rad)
    y = rho * np.sin(az_rad)
    # put the result into a DataFrame
    cart = pd.DataFrame({'x': x, 'y': y, 'z': z})
    return cart


def cart2cyl(x, y, z):
    """function to convert cartesian coordinates to cylindrical coordinates.
    It expects numpy arrays as input with the same lenght or lenght of 1"""
    # do the maths
    rho = np.hypot(x, y)
    az_rad = np.arctan2(x, y)
    # convert to degrees and rotate 90 degrees to have North be 90 degrees
    az_deg = -np.rad2deg(az_rad) + 90
    # wrap around the parts above 360 due to rotate
    az_deg[az_deg < 0] = az_deg[az_deg < 0] + 360
    # put the result into a DataFrame
    cyl = pd.DataFrame({'azimuth': az_deg, 'z': z, 'rho': rho})
    return cyl


def xyz_coords_from_sps(sps):
    """Takes sps dict as input. Extracts x, y, z coords from an sps dict made
    with the create_sensor_points() or the pointgrid_over_surface() function"""
    x, y, z = ([] for i in list(range(3)))
    for p in sps:
        x += [c[0] for c in p['sensor_points']]
        y += [c[1] for c in p['sensor_points']]
        z += [c[2] for c in p['sensor_points']]
    return pd.DataFrame(data={'x':x, 'y':y, 'z':z})


def xyz2proj(xyz, testpoint_height, cull_tol=1.5, plot_proj=False,
             save_projplot=False, save_proj=False, proj_name='proj', cyl=False,
             radius=1):
    """Function to go from an xyz pointcloud file around the testpoint,
    to a projected sphere in 3D with the ability to plot it and
    return the x, y, z DataFrame and the azimuth, z, rho DataFrame (in case of
    cyl = True) or the azimuth, elevation, radius DataFrame (in case of
    cyl = False).
    xyz = xyz-array (created with tiff2xyz function)
    testpoint_height = z-value of testpoint (above Amsterdam Ordnance Datum in
        case of tiff-file from pdok.nl
    cull_tol = tolerance which to use for the culling of points that are close
        to one another (in degrees of the azimuth-elevation domain)
    plot_proj = boolean whether to plot the projected points
    save_projplot = boolean whether to save the projected points plot
    save_proj = boolean whether to save the DataFrame of projected points
    proj_name = string with name to use for saving purposes
    cyl = boolean wheter to plot a cylindrical shape (True) or hemispherical
        (False)
    radius = radius of the cylinder / hemisphere."""
    # move LiDAR xyz down so origin = testpoint
    xyz['z'] = xyz['z'] - testpoint_height
    # remove LiDAR xyz points lower than the testpoint
    xyz = xyz[xyz['z'] > 0]
    # convert points to spherical coordinates
    sph = cart2sph(xyz['x'], xyz['y'], xyz['z'])
    # set the radius so the points are 'projected' onto a sphere
    sph['r'] = radius
    if cull_tol > 0:
        # remove points that are close to one another on the sky hemisphere
        sph = remove_neighbors(sph[['azimuth', 'elevation']], tol=cull_tol)
        # convert purged points back to cartesian coordinates
        sphxyz = sph2cart(sph['azimuth'], sph['elevation'], r=radius)
    # if you want to have the points on a cylinder:
    if cyl:
        # create z based on elevation
        z = np.tan(np.deg2rad(sph['elevation'])) * radius
        # convert cartesian to cyl coordinates with the new z
        cyl = cart2cyl(sphxyz['x'], sphxyz['y'], z)
        if plot_proj:
            # convert points back to cartesian coordinates while setting rho to
            # cyl_radius, so they're 'projected' on the cylinder
            cylxyz = cyl2cart(cyl['azimuth'], cyl['z'], rho=radius)
            # plot projected points in 3D
            plotprojected(cylxyz, save=save_projplot)
        # set the return variable
        proj = cyl
    else:
        if plot_proj:
            # plot projected points in 3D
            plotprojected(sphxyz, save=save_projplot)
        # set the return variable
        proj = sph
    if save_proj:
        save_df(proj, filename=proj_name, folder='Projection output')
    return proj



"""
This module contains functions related to calculating flux-transfer
coefficients (sometimes called daylight coefficients).
"""
import pandas as pd
import numpy as np
import time
from matplotlib import pyplot as plt
import matplotlib.colors
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from sklearn.svm import SVC
from sklearn import preprocessing
# Pyrano imports
from pyrano.pointcloud import sph2cart
from pyrano.pointcloud import xyz2proj


def divide_sky(m):
    '''Returns a Pandas DataFrame with the lower, upper, left and right limits
    of the sky-segments. The basis is the Reinhart sky division. m parameter
    controls the further discretization of the Reinhart sky division. It works
    the same way as MF in radiance: m=1 gives the Reinhart sky with 145
    sky-patches (the middle of sky segment 1 points to 360 degs (North), sky
    segment 2 is the one next to it in the counter-clockwise direction); m=2
    divides each sky segment to 4 smaller sky segments except for the top one
    and results in 577 sky-patches. m=4 results in 2305 sky-patches etc.
    Segment 0 is the segment of the ground.'''
    rein_bands = np.linspace(0, 90-(84/7/m/2), 7*m, endpoint=False).tolist()
    rein_top = np.linspace(90-(84/7/m/2), 90, 1*m, endpoint=False).tolist()
    el_lims = rein_bands + rein_top + [90]
    n_az_divs = [30*m]*2*m + [24*m]*2*m + [18*m]*m + [12*m]*m + [6*m]*m + [1]
    el_low_lims = []
    el_high_lims = []
    az_left_lims = []
    az_right_lims = []
    for i in list(range(len(n_az_divs))):
        el_low_lim = el_lims[i]
        el_high_lim = el_lims[i+1]
        az_div = n_az_divs[i]
        az_right_lim = np.linspace(0 + 360/az_div/2, 360 + 360/az_div/2,
                                   az_div, endpoint=False)
        az_left_lim = az_right_lim - 360/az_div
        el_low_lims += [el_low_lim] * len(az_right_lim)
        el_high_lims += [el_high_lim] * len(az_right_lim)
        az_left_lims += az_left_lim.tolist()
        az_right_lims += az_right_lim.tolist()
    sky_seg_lims = pd.DataFrame(data={'sky_patch':range(1, len(el_low_lims) + 1),
                                      'elev_low_lims':el_low_lims,
                                      'elev_high_lims':el_high_lims,
                                      'azim_left_lims':az_left_lims,
                                      'azim_right_lims':az_right_lims})
    sky_seg_lims.set_index('sky_patch', inplace=True)
    # fixing negative azimuths
    sky_seg_lims['azim_left_lims'] += (sky_seg_lims['azim_left_lims'] < 0)*360
    sky_seg_lims['azim_right_lims'] += (sky_seg_lims['azim_right_lims'] < 0)*360
    last_seg = sky_seg_lims.index[-1]
    # suspicious fix for the last new
    sky_seg_lims.loc[last_seg, 'elev_high_lims'] = 90
    # setting ground sky segment
    sky_seg_lims.loc[0] = {'elev_low_lims':-180, 'elev_high_lims':0,
                           'azim_left_lims':0, 'azim_right_lims':360}
    sky_seg_lims.sort_index(inplace=True)
    return sky_seg_lims


def plot_sky_segments(sky_segments, dir_arrows=True, return_ax=False,
                      c_values=None, c_v_lims=None, cmap=None, r=1, alpha=0.2):
    '''Plots the sky-segments in 3D from a Pandas DataFrame made with
    divide_sky function. If return_ax is true then instead of plotting the
    function returns a matplotlib.axes object so further plots can be added to
    it. If dir_arrows is True, arrows for the 4 cardinal directions are added
    to the plot. North arrow is red, east arrow is yellow. c_values = iterable
    to color the sky segments (e.g. with DC values) if = None, the plot uses a
    default color; c_v_lims = tuple or list of 2 values for min and max range
    of coloring with c_values; cmap = matplotlib.cm colormapobject to use for
    coloring the sky segments; r = radius of the 3D sky segment plot; alpha =
    transparency of the colors in the plot.'''
    fig = plt.figure(figsize=(6,6))
    ax = Axes3D(fig, proj_type='persp')
    axlim = r/2 + 0.1*r
    ax.set_xlim(-axlim, axlim)
    ax.set_ylim(-axlim, axlim)
    ax.set_zlim(-axlim, axlim)
    # adding polygons of each sky segment to the plot: the vertices
    # (looking from the inside) start from bottom left and go counter clockwise
    if c_values is not None:
        norm = matplotlib.colors.Normalize(c_v_lims[0], c_v_lims[1])
        colors = [cmap(norm(c_value)) for c_value in c_values]
    else:
        colors = ['orange'] + ['deepskyblue'] * (len(sky_segments) - 1)
    sky_segments['colors'] = colors
    coll3ds = []
    for ss in sky_segments.index[0:-1]:
        elevations = [sky_segments.loc[ss, 'elev_low_lims']]*2
        elevations += [sky_segments.loc[ss, 'elev_high_lims']]*2
        azimuths = [sky_segments.loc[ss, 'azim_left_lims']]
        azimuths += [sky_segments.loc[ss, 'azim_right_lims']]*2
        azimuths += [sky_segments.loc[ss, 'azim_left_lims']]
        ss_coords = sph2cart(az_deg=azimuths, el_deg=elevations, r=r)
        verts = [list(zip(ss_coords['x'], ss_coords['y'], ss_coords['z']))]
        coll3d = Poly3DCollection(verts, alpha=alpha,
                         facecolor=sky_segments.loc[ss, 'colors'],
                         edgecolor='k')
        coll3ds.append(coll3d)
        ax.add_collection3d(coll3d)
        # TODO: implement plotting top sky segment later
    # top sky segment:
    top_segments = sky_segments[sky_segments['elev_high_lims'] ==
                   sky_segments.loc[sky_segments.index[-1], 'elev_low_lims']]
    elevations = top_segments['elev_high_lims']
    azimuths = sky_segments['azim_left_lims']
    ss_coords = sph2cart(az_deg=azimuths, el_deg=elevations, r=r)
    verts = [list(zip(ss_coords['x'], ss_coords['y'], ss_coords['z']))]
    coll3d = Poly3DCollection(verts, alpha=alpha,
                facecolor=sky_segments.loc[sky_segments.index[-1], 'colors'],
                edgecolor='k')
    coll3ds.append(coll3d)
    ax.add_collection3d(coll3d)
    if dir_arrows:
        # add North-South arrows
        arrow_azverts = {'n':[345, 15, 0], 's':[165, 195, 180],
                         'w':[255, 285, 270], 'e':[75, 105, 90]}
        arrow_colors = {'n':'r', 's':'b', 'w':'g', 'e':'yellow'}
        for cd in arrow_azverts.keys():
            arrow = sph2cart(az_deg=arrow_azverts[cd], el_deg=[0, 0, 0],
                             r=[1.125*r, 1.125*r, 1.25*r])
            verts = [list(zip(arrow['x'], arrow['y'], arrow['z']))]
            ax.add_collection3d(Poly3DCollection(verts, alpha=0.5,
                                facecolor=arrow_colors[cd], edgecolor='k'))
    if return_ax:
        return ax
    else:
        plt.show(block=False)


def calc_cr(sp, pointcloud, sky_grid, test_points, sky_segments, c=1, gamma=50,
            weight=2, threshold=0.5, cull_tol=1.5, return_test_points=False,
            verbose=False):
    '''Calculates Cover Ratio of a sensor point for each sky segment based on
    LiDAR point cloud input. sp: sensor_point with 'x' 'y' 'z' coordinates.
    sky_grid: a set of sky-points as training input for the SVm made with the
    uniform_grid_sphr() function. test_points: a set of points where we
    predict the sky shadedness with the SVM made with the
    test_points_for_sky_segments() function. sky_segments: sky segments made
    with the divide_sky() function. c, gamma see here:
    https://scikit-learn.org/stable/auto_examples/svm/plot_rbf_parameters.html
    threshold: threshold value to binarize the SVM prediction (shaded or not
    shaded). cull_tol: angle in degrees used to reduce the density of the
    projected LiDAR points. If return_test_points = True, the test point-wise
    result is returned as well. If verbose = true, information about the
    training and prediction will be printed'''
    # move the pointcloud to view it from the place of the irrad sensorpoint
    pointcloud_sp = pd.DataFrame(data={'x':pointcloud['x'] - sp['x'],
                                       'y':pointcloud['y'] - sp['y'],
                                       'z':pointcloud['z'] - sp['z']})
    # cull and project the pointcloud to a sphere
    start = time.time()
    ptcld_sphr_proj = xyz2proj(xyz=pointcloud_sp,
                      testpoint_height=0, cull_tol=cull_tol,
                      plot_proj=False, save_projplot=False, save_proj=False,
                      proj_name='proj', cyl=False, radius=1)
    azs = list(ptcld_sphr_proj['azimuth'])
    elevs = list(ptcld_sphr_proj['elevation'])
    duration = round(time.time() - start, 2)
    if verbose:
        print(f'Pointcloud culling time = {duration} seconds')
    # preprocess the test grid
    test_grid_sky_segments = []
    for i in test_points.index:
        test_grid_sky_segments.append([test_points.loc[i, 'azim'],
                                       test_points.loc[i, 'elev']])
    test_grid_sky_segments = np.array(test_grid_sky_segments)
    # train the SVM model
    clf, scaler = svm_train(x=azs, y=elevs, skygrid=sky_grid, gamma=gamma, c=c,
                            lidarweight=weight, verbose=verbose)
    # use the SVM model to test the shadedness of the test points
    predicted = svm_predict(clf, scaler, testgrid=test_grid_sky_segments,
                            threshold=threshold, verbose=verbose)
    predicted.rename(columns={'x':'azimuth', 'y':'elevation'}, inplace=True)
    # calculate CR:
    test_points['shaded'] = (1 - predicted['shaded']).to_list()
    sf = pd.Series(index=sky_segments.index)
    for i in sky_segments.index:
        sf[i] = (test_points[test_points['segment'] == i].mean()['shaded'])
    cr = 1 - sf
    if return_test_points:
        return cr, test_points
    else:
        return cr


def calc_empty_flux_transf_coeff(sky_segments, sp_az=0, sp_el=90):
    '''Calculates the Flux transfer coefficients* of the sky and ground
    segments from a Pandas DataFrame made with the divide_sky function. These
    are the flux transfer coeffs for an empty scene, without any obstruction
    for a sensor point looking at the azimuth = sp_az and elevation = sp_el
    (in degrees) direction.
    * dE for each sky segment in equation 1 in Tregenza, P.R., 2015. Daylight
    coefficients. https://doi.org/10.1177/096032718301500201'''
    # sky segment limits
    el_low = np.radians(sky_segments['elev_low_lims'])
    el_high = np.radians(sky_segments['elev_high_lims'])
    az_left = np.radians(sky_segments['azim_left_lims'])
    c_360 = sky_segments['azim_right_lims'] <= sky_segments['azim_left_lims']
    az_right_c = sky_segments['azim_right_lims'] + c_360*360
    az_right = np.radians(az_right_c)
    # calculating DC for each sky segment
    t = np.radians(90 - sp_el) # sensor surface tilt in radians
    z = np.radians(sp_az + 180) # sensor surface azimuth in radians # WHY THE +180??
    #separate calcs for horizontal and non-horizontal dc
    if t == 0:
        dc = -0.25*(np.cos(2*el_high) - np.cos(2*el_low))*(az_right - az_left)
    else:
        dc = -0.25*((np.sin(t)*np.sin(z-az_left)*((np.sin(2*el_high) + 2*el_high) - (np.sin(2*el_low) + 2*el_low)) - np.cos(t)*az_left*(np.cos(2*el_high) - np.cos(2*el_low)))
                 - (np.sin(t)*np.sin(z-az_right)*((np.sin(2*el_high) + 2*el_high) - (np.sin(2*el_low) + 2*el_low)) - np.cos(t)*az_right*(np.cos(2*el_high) - np.cos(2*el_low))))
    dc_clip = dc.clip(0)
    dc_clip[0] = -0.5*np.pi*(np.cos(t)-1)
    return dc_clip


def calc_segment_area(sky_segments):
    '''Calculates the segment areas of the sky and ground segments from a
    Pandas DataFrame made with the divide_sky function '''
    # sky segment limits
    el_low = np.radians(sky_segments['elev_low_lims'])
    el_high = np.radians(sky_segments['elev_high_lims'])
    az_left = np.radians(sky_segments['azim_left_lims'])
    c_360 = sky_segments['azim_right_lims'] <= sky_segments['azim_left_lims']
    az_right_c = sky_segments['azim_right_lims'] + c_360*360
    az_right = np.radians(az_right_c)
    # calculating SA for each sky segment
    sa = -(az_left - az_right) * (np.sin(el_high) - np.sin(el_low))
    sa[0] = np.pi*2 # ground segment is always 2 PI
    return sa


def calc_refl_flux_transf_coeff(e_cos, cr, sky_segments, albedo):
    '''Calculates the "reflected flux transfer coefficients" for each sky
    segment. The reflected flux transfer coefficientsare re-distributed in an
    area-weighted way to the sky and ground segments'''
    e_0 = e_cos.mul((1 - cr), axis=0)
    segment_areas = calc_segment_area(sky_segments)
    delta_e = (e_cos - e_0).sum() * (1 - albedo)
    e_ref = (delta_e/segment_areas.sum()) * segment_areas * (1 - cr)
    return e_ref


def calc_refl_flux_transf_coeff_os(dc_0, dc_cos, dc_sf, sky_segments, albedo):
    '''EXPERIMENTL FUNCTION: Calculates the "reflected DC" for each sky
    segment. The reflected DC of each sky segment is distrubuted
    to the sky segments on the opposite half of the sky hemisphere and the
    ground in an area-weighted way'''
    segment_areas = calc_segment_area(sky_segments)
    segment_midpoints = pd.Series(index=sky_segments.index)
    for i in sky_segments.index:
        az_ll = sky_segments.loc[i, 'azim_left_lims']
        az_rl = sky_segments.loc[i, 'azim_right_lims']
        if np.abs(az_ll - az_rl) > 180:
            segment_midpoints[i] = (az_ll + (az_rl+360)) / 2
        else:
            segment_midpoints[i] = (az_ll + az_rl) / 2
    # collecting "opposite segments"for ech segment
    opposite_segments = []
    for i in sky_segments.index:
        cp = segment_midpoints[i]
        cp_ll = cp - 90
        cp_rl = cp + 90
        if cp_ll < 0:
            cp_ll += 360
        if cp_rl > 360:
            cp_rl -= 360
        os = []
        for sm, ii in list(zip(segment_midpoints, segment_midpoints.index)):
            if (cp >= 90 and cp <= 270):
                if ((sm < cp_ll) | (sm > cp_rl)):
                    os.append(ii)
            else:
                if ((sm < cp_ll) & (sm > cp_rl)):
                    os.append(ii)
        opposite_segments.append(os)
    # top and ground segment has all other segments as opposite
    opposite_segments[-1] = list(sky_segments.index[:-1])
    opposite_segments[0] = list(sky_segments.index[1:])
    opposite_segments = pd.Series(index=sky_segments.index, data=opposite_segments)
    # distributing the reflected dc over the opposite segments in an
    # area-weighted way
    delta_dc = (dc_cos - dc_0) * (1 - albedo)
    dc_refs = pd.Series(index=sky_segments.index, data=[float(0)]*len(sky_segments))
    segment_areas[0] = segment_areas[0] / 2 # should we halve the ground segmetn area?
    for i in sky_segments.index:
        osa = segment_areas[opposite_segments[i]].sum()
        dc_ref = delta_dc[i] / osa
        for j in opposite_segments[i]:
            dc_refs[j] = dc_refs[j] + dc_ref * segment_areas[j] * dc_sf[j]
    return dc_refs


def test_points_for_sky_segments(sky_segments, s):
    '''Generate test points for the sky segments. s = further division of sky
    segments with test points'''
    # TODO: test and doublecheck this and divide_sky if it works the same way
    # as sky division with MF in Radiance
    elevs = []
    azims = []
    segs = []
    for segment in sky_segments.index:
        el = sky_segments.loc[segment, 'elev_low_lims']
        eh = sky_segments.loc[segment, 'elev_high_lims']
        al = sky_segments.loc[segment, 'azim_left_lims']
        ar = sky_segments.loc[segment, 'azim_right_lims']
        de = (eh - el)/s
        da = (ar - al)/s
        if da < 0:
            ar += 360
            da = (ar - al)/s
            aps = np.linspace(al + da/2, ar + da/2, s, endpoint=False)
            aps -= 360
        elif da == 0:
            ar += 360
            da = (ar - al)/s
            aps = np.linspace(al + da/2, ar + da/2, s, endpoint=False)
        else:
            aps = np.linspace(al + da/2, ar + da/2, s, endpoint=False)
        eps = np.linspace(el + de/2, eh + de/2, s, endpoint=False)

        for e in eps:
            for a in aps:
                elevs.append(e)
                azims.append(a)
                segs.append(segment)
    test_points = pd.DataFrame(data={'azim':azims, 'elev':elevs,
                                     'segment':segs})
    return test_points


def read_pts_file(sps_file):
    'Reading the sps file to a dataframe'
    pts = pd.read_csv(sps_file, sep=' ', names=['x', 'y', 'z', 'vx', 'vy',
                                                'vz'])
    return pts


def write_dc_file(dc, dc_file_name, ncomp=3):
    '''Writes dc file made with pyrano. If append=True it adds a new DC line to
    an existing file'''
    # TODO: this is just a quick prototype. Make this more general later.
    ncols = dc.shape[0]
    if(len(dc.shape)) == 1:
        nrows = 1
    else:
        nrows = dc.shape[1]
    with open(dc_file_name , 'w', newline='\n') as f:
        f.write('#Pyrano\n')
        f.write('#\n')
        f.write('#\n')
        f.write('#\n')
        f.write('#\n')
        f.write('#\n')
        f.write('NCOMP={}\n'.format(str(ncomp)))
        f.write('NROWS={}\n'.format(str(nrows)))
        f.write('NCOLS={}\n'.format(str(ncols)))
        f.write('FORMAT=ascii\n')
        f.write('\n')
        if(len(dc.shape)) == 1:
            dcstr = ''
            for dc_i in dc:
                dcstr += str(dc_i) + '\t' + str(dc_i) + '\t' + str(dc_i) + '\t'
            f.write(dcstr + '\n')
        else:
            for c in dc.columns:
                dcstr = ''
                for dc_i in dc[c]:
                    dcstr += str(dc_i) + '\t' + str(dc_i) + '\t' + str(dc_i) + '\t'
                f.write(dcstr + '\n')


def uniform_grid_sphr(az_res=15, zen_res=10):
    """Create a grid uniformly in the testspace (azimuth=0-360, elevation=0-90)"""
    elevs = np.linspace(0, 90, zen_res, endpoint=False)
    azs = []
    for elev in elevs:
        elev_rad = np.radians(elev)
        azs.append(list(np.linspace(0, 360, az_res*np.cos(elev_rad), endpoint=False)))
    uniformgrid = []
    for n_row, elev in enumerate(elevs):
        uniformgrid += [[az, elev] for az in azs[n_row]]
    return np.array(uniformgrid)


def svm_train(x=[], y=[], skygrid=[], gamma=5, c=20, lidarweight=1, verbose=False):
    """x = x-values of LiDAR points. y = y-values of LiDAR points
    skygrid = a list of x's and y's where there is sky. Can be generated
    using uniform_grid or grid_reader
    gamma = gamma parameter as documented here: https://scikit-learn.org/stable/auto_examples/svm/plot_rbf_parameters.html (default = 5)
    lidarweight = how many times heavier the LiDAR points should 'weigh'
    compared to the skygrid (default = 1)"""
    # start time of training
    start = time.time()
    # create one DataFrame with the x and y values of the LiDAR points
    lidar = pd.DataFrame({'lidar_x':x, 'lidar_y':y})
    # put the lidar and skygrid DataFrame into one list
    X = np.concatenate((lidar, skygrid))
    # make a list the length of both the LiDAR and skygrid with 1's and 0's, where 1 = LiDAR and 0 = skygrid
    y = [1] * len(lidar) + [0] * len(skygrid)
    # create the SVM RBF function
    clf = SVC(C=c, cache_size=2048, class_weight='balanced', coef0=0.0,
              decision_function_shape=None, degree=3, gamma=gamma, kernel='rbf',
              max_iter=-1, probability=False, random_state=None, shrinking=False,
              tol=0.001, verbose=verbose)
    # scale the X data with the preprocessing scaler
    scaler = preprocessing.StandardScaler().fit(X)
    X_scaled = scaler.transform(X)
    # add desired weight to the LiDAR points
    sample_weight = [lidarweight] * len(lidar) + [1] * len(skygrid)
    # fit the data with desired weights
    clf.fit(X_scaled, y, sample_weight=sample_weight)
    # end the prediction timer and print the time
    duration = round(time.time() - start, 2)
    if verbose:
        print(f'Training time = {duration} seconds')
    return (clf, scaler)


def svm_predict(clf, scaler, testgrid=[], threshold=0.5, verbose=False):
    """clf = SVM model, scaler = scaler used to make the SVM model. Both are
    outputs from svm_train. testgrid = a list of x's and y's where we want
    to test if it's shaded or not. Can be generated using uniform_grid """
    # start time of prediction
    start = time.time()
    # scale the testgrid with the same scaler. Testgrid is aka Xtest
    Xtest_scaled = scaler.transform(testgrid)
    # predict if the points are shaded or not
    ytest = clf.decision_function(Xtest_scaled)
    # binarize the points with a certain threshold between 0 and 1, where 1 is shaded and 0 is not shaded
    ytest[ytest >= threshold] = 1
    ytest[ytest < threshold] = 0
    # create DataFrame with x and y of the points and whether the point is shaded
    try:
        # if the testgrid is a pandas DataFrame
        predicted_values = pd.DataFrame({'x':testgrid['x'], 'y':testgrid['y'],
                                        'shaded':ytest})
    except:
        # if the testgrid is a numpy array
        predicted_values = pd.DataFrame({'x':testgrid[:,0], 'y':testgrid[:,1],
                                        'shaded':ytest})
    # end the prediction timer and print the time
    duration = round(time.time() - start, 2)
    if verbose:
        print(f'Prediction time = {duration} seconds')
    # return the predicted values and the ax from the figure to use in future functions
    return predicted_values



from distutils.core import setup
import os

long_description = '''
Pyrano is a Python package for simulating solar irradiance
on external built surfaces using Digital Surface Model (DSM)
point clouds as shading geometry. Pyrano aims to bridge the gap
between building energy, solar irradiance and PV power simulations.
It is compatible with EnergyPlus, Radiance and PVMismatch. \n
See more at https://gitlab.tue.nl/bp-tue/pyrano
'''

setup(
  name = 'pyrano',
  packages = ['pyrano'],
  version = '0.1.2',
  license = 'MIT',
  description = 'Solar irradiance simulations for PV and BPS applications',
  long_description_content_type = 'text/markdown',
  long_description = long_description,
  author = 'Adam Bognar',
  author_email = 'a.bognar@tue.nl',
  url = 'https://gitlab.tue.nl/bp-tue/pyrano',
  download_url = 'https://gitlab.tue.nl/bp-tue/pyrano/-/archive/v0.1.2/pyrano-v0.1.2.tar.gz',
  keywords = ['solar', 'irradiance', 'pv', 'radiance', 'energyplus', 'pvmismatch', 'lidar', 'pointclouds'],
  install_requires=['pandas==0.25.3',
                    'numpy==1.19.5',
                    'matplotlib==3.1.3',
                    'eppy==0.5.54',
                    'geomeppy==0.11.8',
                    'scikit-learn==0.24.1',
                    'Pillow==8.1.0',
                    'Rtree==0.9.7'],
  classifiers=[
    'License :: OSI Approved :: MIT License',
    'Development Status :: 4 - Beta',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python :: 3',
    'Topic :: Scientific/Engineering',
  ],
)
